// Pathfinder
// Writen by David J. Stein, Esq., June 2005. See "License.html" for terms and conditions of use.
//
// This module contains the front end for the simulator.

#region Using declarations

using System;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;
using System.Data;

#endregion

namespace Pathfinder_NS {

	public class PathfinderSimulator : System.Windows.Forms.Form {

		#region Variables, initialization code, and Main

		private System.Windows.Forms.PictureBox picPathfinder;
		private System.Windows.Forms.Timer timerUpdate;
		private System.ComponentModel.IContainer components;
		private System.Windows.Forms.Label label1;
		private System.Windows.Forms.NumericUpDown nudObstacleCount;
		private System.Windows.Forms.Label label2;
		private System.Windows.Forms.Label label3;
		private System.Windows.Forms.NumericUpDown nudObstacleSize;
		private System.Windows.Forms.Label label4;
		private System.Windows.Forms.TrackBar trackSpeed;
		private System.Windows.Forms.Button btnSolveStep;
		private System.Windows.Forms.Button btnSolveAnimate;
		private System.Windows.Forms.Button btnSolveFull;

		Pathfinder p;
		long lStartX, lStartY, lEndX, lEndY, lTotalDistance;
		CollisionRect CurrentOb, StartR, StartB;
		uint iTotalTests, iTotalCycles, iCycle, iEndOfTestTimer, iFailures;
		int iTimerLength;
		bool bOutput;
		bool bAnimate = false;
		Random r;

		public PathfinderSimulator() {
			InitializeComponent();
			r = new Random((int) System.DateTime.Now.Ticks);
			p = null;
			iTotalTests = iTotalCycles = iEndOfTestTimer = 0;
			lTotalDistance = 0;
			iTimerLength = 2000;
			bOutput = true;
			iFailures = 0;
			Process();
		}

		#region Uninteresting Windows Form Designer generated code

		private void InitializeComponent() {
			this.components = new System.ComponentModel.Container();
			this.picPathfinder = new System.Windows.Forms.PictureBox();
			this.timerUpdate = new System.Windows.Forms.Timer(this.components);
			this.label1 = new System.Windows.Forms.Label();
			this.nudObstacleCount = new System.Windows.Forms.NumericUpDown();
			this.label2 = new System.Windows.Forms.Label();
			this.label3 = new System.Windows.Forms.Label();
			this.nudObstacleSize = new System.Windows.Forms.NumericUpDown();
			this.label4 = new System.Windows.Forms.Label();
			this.trackSpeed = new System.Windows.Forms.TrackBar();
			this.btnSolveStep = new System.Windows.Forms.Button();
			this.btnSolveAnimate = new System.Windows.Forms.Button();
			this.btnSolveFull = new System.Windows.Forms.Button();
			((System.ComponentModel.ISupportInitialize)(this.nudObstacleCount)).BeginInit();
			((System.ComponentModel.ISupportInitialize)(this.nudObstacleSize)).BeginInit();
			((System.ComponentModel.ISupportInitialize)(this.trackSpeed)).BeginInit();
			this.SuspendLayout();
			// 
			// picPathfinder
			// 
			this.picPathfinder.Anchor = ((System.Windows.Forms.AnchorStyles)((((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Bottom) 
				| System.Windows.Forms.AnchorStyles.Left) 
				| System.Windows.Forms.AnchorStyles.Right)));
			this.picPathfinder.BorderStyle = System.Windows.Forms.BorderStyle.FixedSingle;
			this.picPathfinder.Location = new System.Drawing.Point(6, 8);
			this.picPathfinder.Name = "picPathfinder";
			this.picPathfinder.Size = new System.Drawing.Size(316, 320);
			this.picPathfinder.TabIndex = 0;
			this.picPathfinder.TabStop = false;
			this.picPathfinder.Paint += new System.Windows.Forms.PaintEventHandler(this.picPathfinder_Paint);
			// 
			// timerUpdate
			// 
			this.timerUpdate.Enabled = true;
			this.timerUpdate.Interval = 2000;
			this.timerUpdate.Tick += new System.EventHandler(this.timerUpdate_Tick);
			// 
			// label1
			// 
			this.label1.Anchor = System.Windows.Forms.AnchorStyles.Bottom;
			this.label1.Location = new System.Drawing.Point(8, 368);
			this.label1.Name = "label1";
			this.label1.Size = new System.Drawing.Size(64, 23);
			this.label1.TabIndex = 5;
			this.label1.Text = "Obstacles:";
			this.label1.TextAlign = System.Drawing.ContentAlignment.MiddleLeft;
			// 
			// nudObstacleCount
			// 
			this.nudObstacleCount.Anchor = System.Windows.Forms.AnchorStyles.Bottom;
			this.nudObstacleCount.Location = new System.Drawing.Point(152, 368);
			this.nudObstacleCount.Maximum = new System.Decimal(new int[] {
																								 300,
																								 0,
																								 0,
																								 0});
			this.nudObstacleCount.Minimum = new System.Decimal(new int[] {
																								 1,
																								 0,
																								 0,
																								 0});
			this.nudObstacleCount.Name = "nudObstacleCount";
			this.nudObstacleCount.Size = new System.Drawing.Size(56, 20);
			this.nudObstacleCount.TabIndex = 6;
			this.nudObstacleCount.Value = new System.Decimal(new int[] {
																							  40,
																							  0,
																							  0,
																							  0});
			// 
			// label2
			// 
			this.label2.Anchor = System.Windows.Forms.AnchorStyles.Bottom;
			this.label2.Location = new System.Drawing.Point(96, 368);
			this.label2.Name = "label2";
			this.label2.Size = new System.Drawing.Size(48, 23);
			this.label2.TabIndex = 7;
			this.label2.Text = "Number";
			this.label2.TextAlign = System.Drawing.ContentAlignment.MiddleLeft;
			// 
			// label3
			// 
			this.label3.Anchor = System.Windows.Forms.AnchorStyles.Bottom;
			this.label3.Location = new System.Drawing.Point(224, 368);
			this.label3.Name = "label3";
			this.label3.Size = new System.Drawing.Size(32, 23);
			this.label3.TabIndex = 9;
			this.label3.Text = "Size";
			this.label3.TextAlign = System.Drawing.ContentAlignment.MiddleLeft;
			// 
			// nudObstacleSize
			// 
			this.nudObstacleSize.Anchor = System.Windows.Forms.AnchorStyles.Bottom;
			this.nudObstacleSize.Location = new System.Drawing.Point(264, 368);
			this.nudObstacleSize.Maximum = new System.Decimal(new int[] {
																								200,
																								0,
																								0,
																								0});
			this.nudObstacleSize.Minimum = new System.Decimal(new int[] {
																								1,
																								0,
																								0,
																								0});
			this.nudObstacleSize.Name = "nudObstacleSize";
			this.nudObstacleSize.Size = new System.Drawing.Size(56, 20);
			this.nudObstacleSize.TabIndex = 8;
			this.nudObstacleSize.Value = new System.Decimal(new int[] {
																							 40,
																							 0,
																							 0,
																							 0});
			// 
			// label4
			// 
			this.label4.Anchor = System.Windows.Forms.AnchorStyles.Bottom;
			this.label4.Location = new System.Drawing.Point(8, 392);
			this.label4.Name = "label4";
			this.label4.Size = new System.Drawing.Size(72, 23);
			this.label4.TabIndex = 10;
			this.label4.Text = "Step Interval:";
			this.label4.TextAlign = System.Drawing.ContentAlignment.MiddleLeft;
			// 
			// trackSpeed
			// 
			this.trackSpeed.Anchor = System.Windows.Forms.AnchorStyles.Bottom;
			this.trackSpeed.Location = new System.Drawing.Point(80, 392);
			this.trackSpeed.Maximum = 2000;
			this.trackSpeed.Minimum = 1;
			this.trackSpeed.Name = "trackSpeed";
			this.trackSpeed.Size = new System.Drawing.Size(248, 42);
			this.trackSpeed.TabIndex = 11;
			this.trackSpeed.TickFrequency = 100;
			this.trackSpeed.Value = 2000;
			this.trackSpeed.ValueChanged += new System.EventHandler(this.trackSpeed_ValueChanged);
			// 
			// btnSolveStep
			// 
			this.btnSolveStep.Location = new System.Drawing.Point(8, 336);
			this.btnSolveStep.Name = "btnSolveStep";
			this.btnSolveStep.Size = new System.Drawing.Size(96, 23);
			this.btnSolveStep.TabIndex = 12;
			this.btnSolveStep.Text = "Solve (Step)";
			this.btnSolveStep.Click += new System.EventHandler(this.btnSolveStep_Click);
			// 
			// btnSolveAnimate
			// 
			this.btnSolveAnimate.Location = new System.Drawing.Point(116, 336);
			this.btnSolveAnimate.Name = "btnSolveAnimate";
			this.btnSolveAnimate.Size = new System.Drawing.Size(96, 23);
			this.btnSolveAnimate.TabIndex = 13;
			this.btnSolveAnimate.Text = "Solve (Animate)";
			this.btnSolveAnimate.Click += new System.EventHandler(this.btnSolveAnimate_Click);
			// 
			// btnSolveFull
			// 
			this.btnSolveFull.Location = new System.Drawing.Point(224, 336);
			this.btnSolveFull.Name = "btnSolveFull";
			this.btnSolveFull.Size = new System.Drawing.Size(96, 23);
			this.btnSolveFull.TabIndex = 14;
			this.btnSolveFull.Text = "Solve (Full)";
			this.btnSolveFull.Click += new System.EventHandler(this.btnSolveFull_Click);
			// 
			// PathfinderSimulator
			// 
			this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
			this.ClientSize = new System.Drawing.Size(328, 427);
			this.Controls.Add(this.btnSolveFull);
			this.Controls.Add(this.btnSolveAnimate);
			this.Controls.Add(this.btnSolveStep);
			this.Controls.Add(this.trackSpeed);
			this.Controls.Add(this.label4);
			this.Controls.Add(this.label3);
			this.Controls.Add(this.nudObstacleSize);
			this.Controls.Add(this.label2);
			this.Controls.Add(this.nudObstacleCount);
			this.Controls.Add(this.label1);
			this.Controls.Add(this.picPathfinder);
			this.MaximizeBox = false;
			this.MinimizeBox = false;
			this.Name = "PathfinderSimulator";
			this.Text = "Pathfinder Simulator";
			this.Resize += new System.EventHandler(this.PathfinderSimulator_Resize);
			this.Load += new System.EventHandler(this.PathfinderSimulator_Load);
			((System.ComponentModel.ISupportInitialize)(this.nudObstacleCount)).EndInit();
			((System.ComponentModel.ISupportInitialize)(this.nudObstacleSize)).EndInit();
			((System.ComponentModel.ISupportInitialize)(this.trackSpeed)).EndInit();
			this.ResumeLayout(false);

		}
		protected override void Dispose(bool disposing) {
			if (disposing) {
				if (components != null) {
					components.Dispose();
				}
			}
			base.Dispose(disposing);
		}
		#endregion

		[STAThread]
		static void Main() {
			Application.Run(new PathfinderSimulator());
		}

		#endregion

		#region Button-click event handlers

		private void trackSpeed_ValueChanged(object sender, System.EventArgs e) {
			timerUpdate.Stop();
			iTimerLength = timerUpdate.Interval = trackSpeed.Value;
			timerUpdate.Start();
		}

		private void btnSolveStep_Click(object sender, System.EventArgs e) {
			bAnimate = false;
			Process();
			picPathfinder.Refresh();
		}

		private void btnSolveFull_Click(object sender, System.EventArgs e) {
			bAnimate = false;
			p = null;
			Process();
			while ((p != null) && (p.bAnalysisComplete == false))
				Process();
			picPathfinder.Refresh();
		}

		private void btnSolveAnimate_Click(object sender, System.EventArgs e) {
			bAnimate = true;
		}

		#endregion

		#region Other event handlers

		private void PathfinderSimulator_Load(object sender, System.EventArgs e) {
			bAnimate = false;
			p = null;
			Process();
			while (p.bAnalysisComplete == false)
				Process();
			picPathfinder.Refresh();
		}

		private void picPathfinder_Paint(object sender, System.Windows.Forms.PaintEventArgs e) {
			if (bOutput == true) {
				Graphics g = e.Graphics;
				g.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.HighQuality;

				Pen CurrentPen = new Pen(Color.Gray, 1);
				Pen StartPen = new Pen(Color.FromArgb(255, 100, 100), 1);
				Pen EndPen = new Pen(Color.FromArgb(50, 175, 50), 1);
				Pen ActivePen = new Pen(Color.FromArgb(40, 100, 128), 1);
				Pen ProcessPen = new Pen(Color.Red, 1);
				Pen BestPen = new Pen(Color.Blue, 3);
				Brush ObstacleBrush = Brushes.Purple;
				Brush ActiveNodeBrush = new SolidBrush(Color.Blue);
				Brush ProcessNodeBrush = new SolidBrush(Color.Red);
				Font font = new Font("Times New Roman", 10.0f);

				float fXScale = picPathfinder.Width / 530.0f;
				float fYScale = picPathfinder.Height / 530.0f;

				// draw cycle count and test metrics
				// draw timer length and (paused) indicator or seconds

				// draw background
				g.FillRectangle(Brushes.BlanchedAlmond, 0, 0, picPathfinder.Width, picPathfinder.Height);
				// draw start and end markers
				g.DrawEllipse(StartPen, (int) ((lStartX - 7) * fXScale), (int) ((lStartY - 7) * fYScale), (int) ((14) * fXScale), (int) ((14) * fYScale));
				g.DrawEllipse(EndPen, (int) ((lEndX - 7) * fXScale), (int) ((lEndY - 7) * fYScale), (int) ((14) * fXScale), (int) ((14) * fYScale));
				// draw obstacles
				CurrentOb = StartR;
				while (CurrentOb != null) {
					g.FillRectangle(ObstacleBrush, (int) ((CurrentOb.iLeft) * fXScale), (int) ((CurrentOb.iTop) * fYScale), (int) ((CurrentOb.iRight - CurrentOb.iLeft) * fXScale), (int) ((CurrentOb.iBottom - CurrentOb.iTop) * fYScale));
					CurrentOb = CurrentOb.NextR;
				}
				// draw paths
				if (p != null) {
					PathNode TraceNode = p.ActiveStart;
					while ((TraceNode != null) && (TraceNode.lDistance < 999999)) {
						if (TraceNode.Parent != null) {
							g.DrawLine(CurrentPen, (int) ((TraceNode.iX) * fXScale), (int) ((TraceNode.iY) * fYScale), (int) ((TraceNode.Parent.iX) * fXScale), (int) ((TraceNode.Parent.iY) * fYScale));
							if (TraceNode == p.ActiveCurrent) {  // draw highlighted node color here
								if (iEndOfTestTimer > 0)
									g.FillEllipse(ProcessNodeBrush, (int) ((lEndX - 6) * fXScale), (int) ((lEndY - 6) * fYScale), (int) ((12) * fXScale), (int) ((12) * fYScale));
								else
									g.FillEllipse(ProcessNodeBrush, (int) ((TraceNode.iX - 6) * fXScale), (int) ((TraceNode.iY - 6) * fYScale), (int) ((12) * fXScale), (int) ((12) * fYScale));
							}
						}
						TraceNode = TraceNode.ActiveNext;
					}
					if (p.bFoundSolution == true) {  // draw path from goal to start
						TraceNode = p.ActiveGoal;
						while (TraceNode.Parent != null) {
							g.DrawLine(BestPen, (int) ((TraceNode.iX) * fXScale), (int) ((TraceNode.iY) * fYScale), (int) ((TraceNode.Parent.iX) * fXScale), (int) ((TraceNode.Parent.iY) * fYScale));
							TraceNode = TraceNode.Parent;
						}
					}
				}
			}
		}

		private void timerUpdate_Tick(object sender, System.EventArgs e) {
			if (bAnimate == true) {
				Process();
				picPathfinder.Refresh();
			}
		}

		private void PathfinderSimulator_Resize(object sender, System.EventArgs e) {
			picPathfinder.Refresh();
		}

		#endregion

		#region Helper functions

		public void Process() {
			CollisionRect Search, TempR, TempB;

			if (p == null) {  // start a new test - create info
				iEndOfTestTimer = 0;
				lStartX = (510 * r.Next(2)) + 6;
				lStartY = (510 * r.Next(2)) + 6;
				lEndX = 522 - lStartX;
				lEndY = 522 - lStartY;
				TempR = TempB = null;
				for (int loop = 0; loop < nudObstacleCount.Value; loop++) {
					// create all obstacles
					CurrentOb = new CollisionRect();
					CurrentOb.Gateway = null;
					CurrentOb.iLeft = 10 + r.Next(470);
					CurrentOb.iTop = 10 + r.Next(470);
					CurrentOb.iRight = CurrentOb.iLeft + 10 + r.Next((int) nudObstacleSize.Value);
					CurrentOb.iBottom = CurrentOb.iTop + 10 + r.Next((int) nudObstacleSize.Value);
					// insert into right and bottom lists
					Search = TempR;
					if (Search == null)
						TempR = CurrentOb;
					else if (Search.iRight >= CurrentOb.iRight) {
						CurrentOb.NextR = TempR;
						TempR = CurrentOb;
					} else {
						while (true) {
							if (Search.NextR == null) {
								Search.NextR = CurrentOb;
								break;
							} else if (Search.NextR.iRight >= CurrentOb.iRight) {
								CurrentOb.NextR = Search.NextR;
								Search.NextR = CurrentOb;
								break;
							}
							Search = Search.NextR;
						}
					}
					Search = TempB;
					if (Search == null)
						TempB = CurrentOb;
					else if (Search.iBottom >= CurrentOb.iBottom) {
						CurrentOb.NextB = TempB;
						TempB = CurrentOb;
					} else {
						while (true) {
							if (Search.NextB == null) {
								Search.NextB = CurrentOb;
								break;
							} else if (Search.NextB.iBottom >= CurrentOb.iBottom) {
								CurrentOb.NextB = Search.NextB;
								Search.NextB = CurrentOb;
								break;
							}
							Search = Search.NextB;
						}
					}
				}
				StartR = TempR;
				StartB = TempB;
				p = new Pathfinder(lStartX, lStartY, lEndX, lEndY, TempR, TempB, new Location(), 1 + r.Next(8));
				p.SetupList();
				iCycle = 0;
			}
			if (iEndOfTestTimer == 0) {  // still processing
				if (p.Solve() == true) {  // success
					if ((iTimerLength == 0) || (bAnimate == false))
						iEndOfTestTimer = CalcTime(0);
					else
						iEndOfTestTimer = CalcTime(2);
					iTotalTests++;
					iTotalCycles += iCycle;
					if (p.bFoundSolution == true)
						lTotalDistance += p.ActiveGoal.lDistance;
					else
						iFailures++;
				} else
					iCycle++;
			} else {
				if (CalcTime(0) > iEndOfTestTimer) {  // get ready for new test
					iEndOfTestTimer = 0;
					StartR = null;
					StartB = null;
					p = null;
				}
			}
		}

		public uint CalcTime(int offset) {
			return (uint) ((System.DateTime.Now.Day * 86400) + (System.DateTime.Now.Hour * 3600) + (System.DateTime.Now.Minute * 60) + (System.DateTime.Now.Second) + offset);
		}

		#endregion

	}

}
